import os

file_path = os.path.dirname(os.path.abspath(__file__))

IMAGE1 = os.path.join(file_path, 'waves1.jpeg')
IMAGE2 = os.path.join(file_path, 'waves2.jpeg')
IMAGE3 = os.path.join(file_path, 'waves3.jpeg')

DEFAULT_IMAGES = [IMAGE1, IMAGE2, IMAGE3]
