/* Author: Matt Clifford <matt.clifford@bristol.ac.uk> 
   License: BSD 3-Clause License */
   
QWidget {
    background-color: "white";
    color: "black";
}
QPushButton {
    /* font-size: 16px; */
    background-color: "white";
    color: "black";
}
QLineEdit {
    background-color: "white";
    color: "black";
}
