# Author: Matt Clifford <matt.clifford@bristol.ac.uk>
# License: BSD 3-Clause License

from IQM_Vis.ui_wrapper import make_UI
from IQM_Vis.data_handlers.data_api import dataset_holder
from IQM_Vis.transforms import *
from IQM_Vis.metrics import *
from IQM_Vis import examples
from IQM_Vis import utils
from IQM_Vis.version import __version__

# Author info
__author__ = "Matt Clifford"
__email__ = "matt.clifford@bristol.ac.uk"
__license__ = "BSD 3-Clause"
